#ifndef __DOCRA_LEXER_HPP
#define __DOCRA_LEXER_HPP

#include <list>
#include <map>
#include <memory>
#include <ostream>
#include <regex>
#include <string>
#include <set>

#include "fs.hpp"

namespace docra {//docra``+

  struct ExtraInfo {
    enum KIND {
      INVALID = -1,
      YACC
    };

    KIND kind = INVALID;

    ExtraInfo(KIND k) : kind(k) {}
  };

  struct Tag { //+`Tag` Identifier of a documented entity
    std::string prefix; //++`.prefix` Prefix for a `KIND.TAG`
    std::string symbol; //++`.symbol` Identifier for a `KIND.TAG`
    std::string suffix; //++`.suffix` Suffix for a `KIND.TAG`

    //++`.empty`
    bool empty()//;?>
    //-> wether this tag is empty
    {
      return prefix.empty() && symbol.empty() && suffix.empty();
    }
    //++`.erase` Clears out the tag contents
    void erase()//;?>
    {
      prefix.erase();
      symbol.erase();
      suffix.erase();
    }

    //++`.xref_id` Turn a tag into an asciidoc identifier for cross-referencing
    std::string xref_id() const;//?>
  };

  inline std::ostream &operator << (std::ostream &os, const Tag &tag)
  {
    os << tag.prefix << '`' << tag.symbol << '`' << tag.suffix;
    return os;
  }

  Tag operator + (const Tag &lhs, const Tag &rhs);

  //+`Chunk` The low-level input to the generic `Lexer`
  struct Chunk {
    enum KIND { //++`::KIND` The specific kind of `Chunk`
      INVALID = -1, //++`.INVALID` Refers to an unitialized `Chunk`
      CODE,         //++`.CODE` A `Chunk` that contains code
      COMMENT,      //++`.COMMENT` A `Chunk` that contains a comment
      TAG           //++`.TAG` A special `Chunk` containing a pre-parsed tag
    };
    KIND kind; //++`.kind` See `KIND`

    std::string str; //++`.str` The text contained by the `Chunk`
    // This field is filled by the specific low-level lexer

    std::string::const_iterator it;
    //++`.it` Current lexing position within a `Chunk`
    // This iterator is repositioned by the `Lexer` while scanning `str`

    int lineno; //++`.lineno` Line within the input file
    // This field is set by the low-level lexers

    int depth; //++`.depth` Hierarchy level
    // This field is also set by the low-level lexers since they know how to deconstruct hierarchy for each language

    Tag tag; //++`.tag` Only set in pre-parsed tags, the tag will be just copied in the resulting token

    std::unique_ptr<ExtraInfo> extra;

    Chunk();
  };

  //+`Token` A piece of input recognized by the `Lexer`
  struct Token {
    enum KIND {//++`::KIND` The specific kind of `Token`
      CODE,        //++`.CODE` A single line of code
      // A `Chunk` of `Chunk.CODE` kind goes through the `Lexer` without modification
      SCOPE_ENTER, //++`.SCOPE_ENTER` Emitted when entering a scope
      SCOPE_EXIT,  //++`.SCOPE_EXIT` Emitted when leaving a scope
      TAG,         //++`.TAG` A tag identifier, with optional prefix and suffix
      INLINE_PARAM,//++`.INLINE_PARAM` A param identifier, rendered inline
      DOC_STR,     //++`.DOC_STR` Emitted by default for parts of comments
      BLOCK_ENTER, //++`.BLOCK_ENTER` The beginning of an embedded code block
      BLOCK_EXIT,  //++`.BLOCK_EXIT` The end of an embedded code block
      LITERAL,     //++`.LITERAL` An inline literal expression
      EOL          //++`.EOL` End-of-line
    };
    KIND kind;  //++`.kind` The `Kind` of this particular token

    bool boc;        //++`.boc` Beginning of comment
    bool brief;      //++`.brief` If a tag, whether it's followed by a column (unused!)
    Tag tag;         //++`.tag` Identifier tag
    int depth;       //++`.depth` Hierarchy level
    std::string str; //++`.str` Token content (for `KIND.CODE` or `KIND.DOC_STR` )
    int lineno;      //++`.lineno` Line number of token

    std::unique_ptr<ExtraInfo> extra;

    bool empty_str() const;

    Token() : kind(EOL), boc(false), brief(false), depth(0), lineno(0) {}

    Token(const Token &t) : kind(t.kind), boc(t.boc), brief(t.brief), tag(t.tag), depth(t.brief), str(t.str), lineno(t.lineno) {}

    Token(Token &&t) = default;
  };

  typedef std::list<Token> Tokens;

  //+`Lexer` Language-agnostic base class for each lexer
  struct Lexer {
    fs::path path;     //++`.path` ?>Optional path to file being scanned
    FILE *file;        //++`.file` ?>Handle to input file
    // We use stdio's FILE since that's what Flex expects.
    bool eof;          //++`.eof` ?>True at end of input
    // See `lex_eof` .
    void *scanner;     //++`.scanner` ?>Opaque pointer to Flex state
    int depth;         //++`.depth` ?>Current position in hierachy
    int depth_delta;   //++`.depth_delta` ?>Cumulative depth variation to be applied at next eol
    Chunk chunk;       //++`.chunk` ?>Last `Chunk` produced by low-level lexer
    // This can briefly differ from chunk.depth, the lexer will generate SCOPE_XXX tokens until balanced.
    std::smatch match; //++`.match` ?>Result of last regexp match
    Token token;       //++`.token` ?>Last `Token` produced by the lexer

    Lexer();

    virtual ~Lexer();

    int lex_eof(); //++`.lex_eof` ?>Sets the `eof` flag

    virtual void init(const fs::path &path, FILE *file) = 0; //++`.init` ?>
    //>`path` path to file
    //>`file` previously opened file handle

    virtual bool next_chunk() = 0;//++`.next_chunk` ?>Low-level chunking
    //-> 'false' on EOF, 'true' otherwise
    // This function is reimplemented by each low-level lexer.

    bool next_token(); //++`.next_token` Produce a new `Token`
    //-> 'false' at the end of input, 'true' otherwise
    Token&& current_token() { return std::move(token); }

    protected:
      virtual void fini() = 0; //++`.fini` ?>Dual of `init`
  };

  //+`Lexer_Dummy` Dummy lexer to test generic logic only
  struct Lexer_Dummy : public Lexer //?>
  {
    void init(const fs::path &path, FILE *) {}
    bool next_chunk(); //++`::next_chunk` ?>
    void fini() {}
  };

  //+`Lexer_Adoc` Lexer recognizing Asciidoc comment syntax
  struct Lexer_Adoc : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_Adoc(); //++`::dtor` ?>
  };

  //+`Lexer_Hash` Lexer for #-commented languages
  struct Lexer_Hash : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_Hash(); //++`::dtor` ?>
  };

  //+`Lexer_CLike` Lexer for C-like languages
  // This includes C, C++, Java, Javascript, Verilog and so forth...
  struct Lexer_CLike : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_CLike(); //++`::dtor` ?>
  };

  //+`Lexer_Vhdl` Specific lexer for VHDL
  struct Lexer_Vhdl : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_Vhdl(); //++`::dtor` ?>
  };

  //+`Lexer_Ini` Lexer for 'ini-like' configuration files
  struct Lexer_Ini : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_Ini(); //++`::dtor` ?>
  };

  typedef std::map<std::string, std::string> map_string_string;
  typedef std::map<std::string, bool> map_string_bool;
  typedef std::pair<std::string, std::string> string_string;
  typedef std::list<string_string> list_string_string;
  typedef std::set<std::string> set_string;

  //+`Lexer_Yacc` Lexer for Yacc/Bison grammars
  struct Lexer_Yacc : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file); //++`::init` ?>
    bool next_chunk(); //++`::next_chunk` ?>
    void fini(); //++`::fini` ?>
    ~Lexer_Yacc(); //++`::dtor` ?>

    map_string_string token_names; //++`.token_names`
//? Since a terminal symbol can be identified both by an identifier and an arbitrary string, this can be used to find the identifier starting from the string.

    std::string line_buffer;    //++`.line_buffer` Buffer holding current line
    set_string token_ids;       //++`.token_ids` Holds all the token identifiers
    std::string curr_id;        //++`.curr_id` The identifier for the token or rule currently being processed
    std::string curr_token_str; //++`.curr_token_str` If `curr_id` points to a token, holds its (optional) string representation
    list_string_string curr_pattern; //++`.curr_pattern` A pattern is a list of terminals and nonterminals, a rule can match several patterns
    int pattern_num;
    int pattern_lineno;
  };

  struct ExtraInfo_Yacc : public ExtraInfo
  {
    ExtraInfo_Yacc(const list_string_string &p)
    : ExtraInfo(YACC), pattern(p) {}

    const list_string_string pattern;
  };

  //+`Lexer_Xml` Lexer for Xml
  struct Lexer_Xml : public Lexer //?>
  {
    void init(const fs::path &path, FILE *file);
    bool next_chunk();
    void fini();
    ~Lexer_Xml();
  };

} //namespace docra

#endif//__DOCRA_LEXER_HPP
