#include <getopt.h>
#include <iostream>
#include <fstream>
#include <string>

#include "docra.hpp"

static struct option getopt_lopts[] = {
//ug`docra.options`verbmore *-v, --verbose*::
// Emit verbose output. Repeat option to increase verbosity level.
  { "verbose", no_argument, 0, 'v'},
//ug`docra.options`verbless *-V, --less-verbose*::
// Less verbose mode. Repeat option to reduce verbosity level.
  { "less-verbose", no_argument, 0, 'V'},
//ug`docra.options`quiet *-q, --quiet*::
// Quiet mode. Suppress any status output message.
  { "quiet", no_argument, 0, 'q'},
//ug`docra.options`config *-c, --config*=_config_::
// Load settings from _config_ configuration file. If this option is repeated, only the last one will be considered.
  { "config", required_argument, 0, 'c'},
//ug`docra.options`outdir *-o, --outdir*=_outdir_::
// Output generated files to _outdir_.
  { "outdir", required_argument, 0, 'o'},
//ug`docra.options`urlprefix *-u, --urlprefix*=_prefix_::
// Use _prefix_ as prefix when creating hyperlinks to source code.
// This option can be repeated after the first path specification to change its value for the following paths.
  { "urlprefix", required_argument, 0, 'u'},
//ug`docra.options`urlsuffix *-U, --urlsuffix*=_suffix_::
// Use _suffix_ as suffix when creating hyperlinks to source code.
// This option can be repeated after the first path specification to change its value for the following paths.
  { "urlsuffix", required_argument, 0, 'U'},
//ug`docra.options`level *-l, --level*=_level_::
// For each source path specification, discard all but the last _level_ levels of hierarchy in the output.
// This option can be repeated after the first path specification to change its value for the following paths.
  { "level", required_argument, 0, 'l'},
//ug`docra.options`paths *-p, --paths*::
// Just list to stdout the paths of the files that would be parsed and exit.
  { "paths", no_argument, 0, 'p'},
  { 0, 0, 0, 0 }
};

static void usage()
{
  std::cerr << "Usage:" << std::endl
    << "docra [-v] [-V] [-q] [-c <config>] [-o <outdir>] [-u <url_prefix>] [-U <url_suffix>]" << std::endl
    << "  [-l <level>] [<source_path> ...] [-p] [<template.dcrt> ...]" << std::endl
    << std::endl;

//ug`docra.synopsis`
// *docra* [-v] [-V] [-q] [-c _config_] [-o _outdir_] [-u _url_prefix_] [-U _url_suffix_]
//   [-l _level_] [_sourcepath_ ...] [-p] [_template.dcrt_ ...]
}

//`main` Entry point of the program
int main(int argc, char *argv[])//;?>
//>`argc` Number of command-line arguments
//>`argv` String array of arguments
{
  docra::Config cfg;
  docra::Parser_Config cfgparser(cfg);
  docra::Project project(cfg);

  //ug`config`0 By default, docra will search for a configuration file called ``Docrafile`` or ``.docra`` in the current working directory.
  if (fs::exists("Docrafile")) {
    cfgparser.parse("Docrafile");
  } else if (fs::exists(".docra")) {
    cfgparser.parse(".docra");
  }

  int getopt_val, optind_curr = 1;
  do {
    getopt_val = getopt_long_only(argc, argv, "vVqc:o:u:U:l:ph", getopt_lopts, nullptr);
    // Process non-options that getopt might have skipped over
    // (that's because we want to process them in order and not all at the end)
    for (; optind_curr < argc; ++optind_curr) {
      if (argv[optind_curr][0] == '-') break;
      std::string optstr = argv[optind_curr];
      fs::path optpath = optstr;
//ug`docra.options`sourcepath _ _sourcepath_::
// Either a folder or a regular file, to be scanned in addition to the paths included by the configuration file.
//ug`docra.options`template _ _template.dcrt_::
// Extra template to be rendered in addition to those specified by the configuration file.
      if (fs::is_directory(optpath)) {
//ug`config`1 If additional directories are given on the command line, they will be interpreted as source paths and docra will try to load the default configuration file from there as well before trying to analyze the directory contents.
        if (cfg.base_dir.empty()) {
          fs::path cfgpath1 = optpath / "Docrafile";
          fs::path cfgpath2 = optpath / ".docra";
          if (fs::exists(cfgpath1)) {
            if (cfg.verbose() >= docra::Config::VERB_INFO) {
              std::cout << "Configuration file: " << cfgpath1 << std::endl;
            }
            cfgparser.parse(cfgpath1);
          } else if (fs::exists(cfgpath2)) {
            if (cfg.verbose() >= docra::Config::VERB_INFO) {
              std::cout << "Configuration file: " << cfgpath2 << std::endl;
            }
            cfgparser.parse(cfgpath2);
          } else {
            if (cfg.verbose() >= docra::Config::VERB_INFO) {
              std::cout << "Source path: " << optpath << std::endl;
            }
            cfg.set("path", optstr, "verbose", cfg.cli_verbose, true);
            cfg.set("path", optstr, "level", cfg.cli_level, true);
          }
        } else {
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Source path: " << optpath << std::endl;
          }
          cfg.set("path", optstr, "verbose", cfg.cli_verbose, true);
          cfg.set("path", optstr, "level", cfg.cli_level, true);
        }
      } else if (fs::is_regular_file(optpath)) {
        if (optpath.extension() == ".dcrt") {
          // It's a template
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Template file: " << optpath << std::endl;
          }
          cfg.set("template", optstr, "verbose", cfg.cli_verbose, true);
          cfg.set("template", optstr, "outdir", cfg.cli_outdir, true);
        } else {
          //It's just a source file
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Source file: " << optpath << std::endl;
          }
          cfg.set("path", optstr, "verbose", cfg.cli_verbose, true);
          cfg.set("path", optstr, "level", cfg.cli_level, true);
        }
      } else {
        std::cerr << "No such file or directory: " << argv[optind_curr] << std::endl;
        return -1;
      }
    }
    if (getopt_val > 0) {
      // A flag was set and we have to find which one
      switch (getopt_val) {
//ug`config.syntax`common.verbose.1 _ The verbosity level can be further increased or decreased from the command line, using the *-v*, *-V* and *-q* parameters.
        case 'v':
          cfg.cli_verbose = cfg.verbose() + 1;
          break;
        case 'V':
          cfg.cli_verbose = cfg.verbose() - 1;
          if (cfg.cli_verbose < 0) {
            cfg.cli_verbose = 0;
          }
          break;
        case 'q':
          cfg.cli_verbose = 0;
          break;
        case 'c':
//ug`config`2 The location of the configuration file can be overridden with the *-c* (*--config*) command line option.
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Configuration file: " << optarg << std::endl;
          }
          cfgparser.parse(optarg);
          break;
        case 'o':
          cfg.cli_outdir = optarg;
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Output directory: " << cfg.cli_outdir << std::endl;
          }
          break;
        case 'u':
          cfg.cli_urlprefix = optarg;
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Url prefix: " << cfg.cli_urlprefix << std::endl;
          }
          break;
        case 'U':
          cfg.cli_urlsuffix = optarg;
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Url suffix: " << cfg.cli_urlsuffix << std::endl;
          }
          break;
        case 'l':
          cfg.cli_level = atoi(optarg);
          if (cfg.verbose() >= docra::Config::VERB_INFO) {
            std::cout << "Level: " << cfg.cli_level << std::endl;
          }
          break;
        case 'p':
          cfg.cli_pathsonly = true;
          break;
        case 'h':
        default:
          usage();
          return -1;
      }
      optind_curr = optind; // so that the for loop starts at optind
    }
  } while(getopt_val >= 0 && optind_curr < argc);

  if (cfg.base_dir.empty()) {
    cfg.base_dir = fs::current_path();
  }

//ug`config.syntax`path.1 By default, if no such section is present, the entire source directory is traversed. If at least one such section is present, only the paths explicitly included in the configuration will be traversed.
  if (!cfg.has_section("path")) {
    cfg.set("path", "./");
  }

  cfg.apply();

  for (auto &path : cfg.section("path")) {
    fs::path path_in = cfg.fullpath(path.first);
    int path_verbose = cfg.verbose("path", path.first);
    int path_level = cfg.path_level(path.first);
    const std::string &url = cfg.path_url(path.first);
    if (fs::is_directory(path_in)) {
      if (cfg.verbose() >= docra::Config::VERB_INFO) {
        std::cout << "Parsing directory: " << path.first << std::endl;
      }
      project.analyze_directory(path_verbose, path_in, path_level);
    } else if (fs::is_regular_file(path_in)) {
      if (cfg.verbose() >= docra::Config::VERB_INFO) {
        std::cout << "Parsing file: " << path.first << std::endl;
      }
      project.analyze_file(path_verbose, path_in, path_level);
    } else {
      std::cout << "Invalid path: " << path.first << std::endl;
    }
  }

  if (cfg.cli_pathsonly) {
    return 0;
  }

  for (auto &tpl : cfg.section("template")) {
    int tpl_verbose = cfg.verbose("template", tpl.first);
    fs::path tpl_in = cfg.fullpath(tpl.first);
    fs::path tpl_outdir = cfg.template_outdir(tpl.first);
    fs::path tpl_outfile = cfg.template_outfile(tpl.first);

    fs::path tpl_out = tpl_outdir / tpl_outfile;

    if (cfg.verbose() >= docra::Config::VERB_INFO) {
      std::cout << "Rendering template " << tpl.first << " to " << tpl_out.string() << std::endl;
    }
    FILE *pfile = fopen(tpl_in.string().c_str(), "r");
    if (pfile == nullptr) {
      std::cerr << "unable to open file " << tpl.first << std::endl;
      continue;
    }
    std::ofstream os(tpl_out);
    if (os.bad()) {
      std::cerr << "unable to write file " << tpl_out.string() << std::endl;
      continue;
    }
    docra::Lexer_Adoc lexer;
    lexer.init(tpl.first, pfile);
    docra::Parser_Template parser(lexer, project, os);
    parser.parse(tpl_verbose);
  }
  if (cfg.section("template").empty()) {
    project.debug();
  }
  return 0;

//ug`docra.exit`
// Non-zero in case of invalid configuration and zero otherwise.
noconfig:
  std::cerr << "Invalid configuration or missing configuration file" << std::endl;
  return -1;
}
