#include "parser.hpp"
#include "project.hpp"
#include "util.hpp"

#include <regex>
#include <iostream>


//docra`Config`+

docra::Config::Config()
: cli_verbose(-1)
, cli_level(-1)
, cli_pathsonly(false)
, cli_urlsuffix("#L")
{
  reset();
}

void docra::Config::reset()
{
  base_dir.clear();
  sections.clear();
}

//ug`config.syntax`lexer _ ==== [lexer "type"] section
// This section can be used to control how docra decides what lexer to use to analyze a given file. The following variables are available:
// _ include = _pattern_::
// Files matching this pattern will be scanned by the given lexer. If multiple lexers accept the same patterns the result is undefined.
// _ exclude = _pattern_::
// Files matching this pattern will not be scanned by this lexer. Another lexer may accept this pattern instead.
// _ The following lexer types are available:

void docra::Config::exclude(Filters &filters, const std::string &spec)
{
  filters.emplace_back(std::make_pair(false, make_regex(spec, true)));
}

void docra::Config::include(Filters &filters, const std::string &spec)
{
  filters.emplace_back(std::make_pair(true, make_regex(spec, true)));
}

void docra::Config::apply(const std::string &sec, const std::string &sub, Filters& filters)
{
  auto &entries = get(sec, sub);
  for (auto &entry : entries) {
    if (entry.first == "include") {
      include(filters, entry.second);
    } else if (entry.first == "exclude") {
      exclude(filters, entry.second);
    }
  }
}

void docra::Config::apply()
{
  apply("common", "", common_filters);
  apply("lexer", "clike", lex_clike_filters);
  apply("lexer", "hash", lex_hash_filters);
  apply("lexer", "vhdl", lex_vhdl_filters);
  apply("lexer", "ini", lex_ini_filters);
  apply("lexer", "yacc", lex_yacc_filters);
  apply("lexer", "xml", lex_xml_filters);

//ug`config.syntax`common.exclude.1 Several rules are defined by default to exclude unrecognized file formats, hidden ("dot") folders or paths that do not contain source code.
  exclude(common_filters, "$^.*/?[.].+/.*$");
  exclude(common_filters, "$^.*/?node_modules/$");
  exclude(common_filters, "*.png");
  exclude(common_filters, "*.jpg");
  exclude(common_filters, "*.pdf");
  exclude(common_filters, "*.xlsx");
  exclude(common_filters, "*.deb");
  exclude(common_filters, "*.rpm");
  exclude(common_filters, "*.zip");
  exclude(common_filters, "*.swp");
  exclude(common_filters, "*.a");
  exclude(common_filters, "*.o");
  exclude(common_filters, "*.ko");
  exclude(common_filters, "*.so");

//ug`config.syntax`lexer.clike.1 This lexer is used by default on C, C++, Verilog, SystemVerilog, Javascript, JSON, Java, Kotlin and Groovy files.
  include(lex_clike_filters, "*.c");
  include(lex_clike_filters, "*.h");
  include(lex_clike_filters, "*.cpp");
  include(lex_clike_filters, "*.hpp");
  include(lex_clike_filters, "*.lpp");
  include(lex_clike_filters, "*.v");
  include(lex_clike_filters, "*.sv");
  include(lex_clike_filters, "*.js");
  include(lex_clike_filters, "*.json");
  include(lex_clike_filters, "*.java");
  include(lex_clike_filters, "*.kt");
  include(lex_clike_filters, "*.gradle");

//ug`config.syntax`lexer.hash.1 This lexer is used by default on CMake, shell, YAML, TCL, make and docker files.
  include(lex_hash_filters, "$^.*/?CMakeLists[.]txt$");
  include(lex_hash_filters, "*.cmake");
  include(lex_hash_filters, "*.sh");
  include(lex_hash_filters, "*.yml");
  include(lex_hash_filters, "*.tcl");
  include(lex_hash_filters, "$^.*/?Makefile$");
  include(lex_hash_filters, "$^.*/?GNUMakefile$");
  include(lex_hash_filters, "$^.*/?Dockerfile$");

  include(lex_vhdl_filters, "*.vhd");
  include(lex_vhdl_filters, "*.vhdl");

//ug`config.syntax`lexer.ini.1 This lexer is used by default on INI and docra configuration files.
  include(lex_ini_filters, "*.ini");
  include(lex_ini_filters, "$^.*/?[.]docra$");
  include(lex_ini_filters, "$^.*/?Docrafile$");

//ug`config.syntax`lexer.adoc.1 This lexer is used by default on AsciiDoc files.
  include(lex_adoc_filters, "*.adoc");

//ug`config.syntax`lexer.yacc.1 This lexer is used by default on ``.y`` and ``.ypp`` files (Flex scanners use the ``clike`` lexer instead).
  include(lex_yacc_filters, "*.y");
  include(lex_yacc_filters, "*.ypp");

//ug`config.syntax`lexer.xml.1 This lexer is used by default on XML files.
  include(lex_xml_filters, "*.xml");
}

docra::ConfigEntries::reverse_iterator docra::Config::find(ConfigEntries &entries, const std::string &str)
{
  for (auto i = entries.rbegin(); i != entries.rend(); ++i) {
    if (i->first == str) return i;
  }
  return entries.rend();
}

//ug`config.syntax`common _ ==== [common] section
// This section defines global parameters that are applied in addition to the internal defaults, the following variables are recognized in this section:

//++`::verbose`
int docra::Config::verbose(const std::string &sec, const std::string &sub)
{
  if (cli_verbose >= 0) {
    return cli_verbose;
  }
  auto secit = sections.find(sec);
  if (secit == sections.end()) {
common:
    secit = sections.find("common");
    if (secit == sections.end()) {
      goto nocfg;
    }
    auto subit = secit->second.find("");
    if (subit == secit->second.end()) {
      goto nocfg;
    }
    auto keyit = find(subit->second, "verbose");
    if (keyit == subit->second.rend()) {
      goto nocfg;
    }
    return atoi(keyit->second.c_str());
  } else {
    auto subit = secit->second.find(sub);
    if (subit == secit->second.end()) {
      goto common;
    }
    auto keyit = find(subit->second, "verbose");
    if (keyit == subit->second.rend()) {
      goto common;
    }
    return atoi(keyit->second.c_str());
  }
nocfg:
  return VERB_INFO;
}

//ug`config.syntax`path _ ==== [path "relative/path/"] section
// A path section can be used to tell docra to traverse specific paths.

//ug`config.syntax`path.2 Allowed variables are:
//ug`config.syntax`path.level _ level::
// The nesting level that this path starts at when traversing. This means that only the last _level_ elements of the path will be used when constructing links to source code. This parameter defaults to the length of the path which should be sufficient in most cases.

int docra::Config::path_level(const fs::path &sub)
{
  auto secit = sections.find("path");
  if (secit == sections.end()) {
common:
    if (cli_level >= 0) {
      return cli_level;
    }
    secit = sections.find("common");
    if (secit == sections.end()) {
      goto nocfg;
    }
    auto subit = secit->second.find("");
    if (subit == secit->second.end()) {
      goto nocfg;
    }
    auto keyit = find(subit->second, "level");
    if (keyit == subit->second.rend()) {
      goto nocfg;
    }
    return atoi(keyit->second.c_str());
  } else {
    auto subit = secit->second.find(sub.string());
    if (subit == secit->second.end()) {
      goto common;
    }
    auto keyit = find(subit->second, "level");
    if (keyit == subit->second.rend()) {
      goto common;
    }
    return atoi(keyit->second.c_str());
  }
nocfg:
  int ret = 0;
  for (auto pe : sub) if (pe != ".") ++ret;
  return ret;
}

////ug`config.syntax`path.url url::
//// The URL prefix to use when generating links to source code under this path.

std::string docra::Config::path_url(const fs::path &sub)
{
  auto secit = sections.find("path");
  if (secit == sections.end()) {
common:
    if (!cli_urlprefix.empty()) {
      return cli_urlprefix;
    }
    secit = sections.find("common");
    if (secit == sections.end()) {
      goto nocfg;
    }
    auto subit = secit->second.find("");
    if (subit == secit->second.end()) {
      goto nocfg;
    }
    auto keyit = find(subit->second, "url");
    if (keyit == subit->second.rend()) {
      goto nocfg;
    }
    return keyit->second.c_str();
  } else {
    auto subit = secit->second.find(sub.string());
    if (subit == secit->second.end()) {
      goto common;
    }
    auto keyit = find(subit->second, "url");
    if (keyit == subit->second.rend()) {
      goto common;
    }
    return keyit->second;
  }
nocfg:
  return "";
}

//ug`config.syntax`common.outdir _ outdir = _output/path_::
// Sets the directory where the AsciiDoc files should be generated. The current working directory is used by default. The *-o* command line option has higher precedence than this parameter and can be used to override it.

//ug`config.syntax`template _ ==== [template "file/path.dcrt"] section
// Such a section indicates to docra to load the given file as an AsciiDoc template. These templates will be converted to AsciiDoc files after docra has collected the documentation tags from the entire source code tree.

//ug`config.syntax`template.outdir _ outdir = _output/path_::
// This variable has the same meaning as the common parameter of the same name, it can be used to override the output directory for a specific template.

fs::path docra::Config::template_outdir(const fs::path &sub)
{
  auto secit = sections.find("template");
  if (secit == sections.end()) {
common:
    if (!cli_outdir.empty()) {
      return cli_outdir;
    }
    secit = sections.find("common");
    if (secit == sections.end()) {
      goto nocfg;
    }
    auto subit = secit->second.find("");
    if (subit == secit->second.end()) {
      goto nocfg;
    }
    auto keyit = find(subit->second, "outdir");
    if (keyit == subit->second.rend()) {
      goto nocfg;
    }
    return fullpath(keyit->second);
  } else {
    auto subit = secit->second.find(sub.string());
    if (subit == secit->second.end()) {
      goto common;
    }
    auto keyit = find(subit->second, "outdir");
    if (keyit == subit->second.rend()) {
      goto common;
    }
    return keyit->second;
  }
nocfg:
  return ".";
}

fs::path docra::Config::template_outfile(const fs::path &sub)
{
  fs::path outfile = sub.filename();
  outfile.replace_extension(".adoc");
  auto secit = sections.find("template");
  if (secit == sections.end()) {
    return outfile;
  } else {
    auto subit = secit->second.find(sub.string());
    if (subit == secit->second.end()) {
      return outfile;
    }
    auto keyit = find(subit->second, "outfile");
    if (keyit == subit->second.rend()) {
      return outfile;
    }
    return keyit->second;
  }
}

void docra::Config::set(const std::string &sec, const std::string &sub)
{
  // Just create an empty subsection if it does not exist
  auto &secel = sections[sec];
  auto &subel = secel[sub];
}

//++`::set`
void docra::Config::set(const std::string &sec, const std::string &sub, const std::string &key, const std::string &val, bool replace)
{
  auto &secel = sections[sec];
  auto &subel = secel[sub];
  if (replace) {
    // If key already exists, update its value, don't add a new one
    auto keyit = find(subel, key);
    if (keyit != subel.rend()) {
      keyit->second = val;
      return;
    }
  }
  subel.emplace_back(key, val);
}

//++`::get`
const docra::ConfigEntries &docra::Config::get(const std::string &sec, const std::string &sub)
{
  auto secem = sections.emplace(sec, ConfigSubSections{});
  auto subem = secem.first->second.emplace(sub, ConfigEntries{});
  return subem.first->second;
}

//docra`Scope`+

//++`::add_child` Creates a new `Scope` inside the current one
docra::Scope* docra::Scope::add_child(int lineno)
{
  children.emplace_back();
  Scope *child = &children.back();
  child->parent = this;
  child->lineno_start = lineno;
  return child;
}

void docra::Scope::print(size_t level)
{
  if (!suffix) {
    return;
  }
  std::cerr << std::string(level * 2, ' ');
  std::cerr << suffix->tag.prefix << '`' << suffix->tag.symbol << '`' << suffix->tag.suffix << std::endl;
  std::cerr << std::string(level * 2, ' ');
  std::cerr << "{" << std::endl;
  for (auto &s : children) {
    s.print(level + 1);
  }
  std::cerr << std::string(level * 2, ' ');
  std::cerr << "}" << std::endl;
}

static int expand_tag(const docra::Tag &parent, docra::Tag &tag)
{
  int leveloffset = 0;
  if (tag.prefix == "++") {
    leveloffset = +1;
  }
  if (tag.prefix == "+" || tag.prefix == "++") {
    tag.prefix = parent.prefix;
    tag.symbol.insert(0, parent.symbol);
  }
  return leveloffset;
}

static void error(const fs::path &path, const docra::Token &tok, const std::string &msg)
{
  std::cerr << path.string() << ":" << tok.lineno << ":1: error: " << msg << std::endl;
}

//docra`Parser_Source`+

docra::Parser_Source::Parser_Source(docra::Lexer &l, docra::Project &p)
//docra`Parser_Source`
//>`l` a `Lexer`
//>`p` a `Project`
: lexer(l), project(p)
{}

void docra::Param::render(const Config &cfg, const Suffix &suffix, std::ostream& out) const
{
  // TODO: also generate a cross-referenceable label so that I can link to this param from elsewhere in the document
  out << std::endl;
  switch (kind) {
    case Param::IN:
      out << name << " _(I)_::" << std::endl;
      break;
    case Param::OUT:
      out << name << " _(O)_::" << std::endl;
      break;
    case Param::INOUT:
      out << name << " _(IO)_::" << std::endl;
      break;
    case Param::RET:
      out << "Returns::" << std::endl;
      break;
    default:
      out << name << "::" << std::endl;
      break;
  }
  if (!desc_code.empty()) {
    Tokens::const_iterator ci = desc_code.begin();
    //TODO: maybe this needs a list continuation (a plus sign on a line by itself)
    out << "+";
    suffix.render_codeblock(out, desc_code, ci);
  }
  out << "+" << std::endl;
  suffix.render_description(cfg, out, desc);
  out << std::endl << std::endl;
}

static docra::Param *make_param(docra::Suffix *suffix, const docra::Token &token)
{
  using namespace docra;
  if (!suffix) {
    return nullptr;
  }
  Param::KIND kind;
  if (token.tag.prefix == ">") {
    kind = Param::IN;
  } else if (token.tag.prefix == "<") {
    kind = Param::OUT;
  } else if (token.tag.prefix == "><") {
    kind = Param::INOUT;
  } else if (token.tag.prefix == "=") {
    kind = Param::GENERIC;
  } else if (token.tag.prefix == "-") {
    kind = Param::LOCAL;
  } else {
    return nullptr;
  }
  suffix->params.emplace_back(kind, token.tag.symbol, token.lineno);
  return &suffix->params.back();
}

//++`::parse` Parse the current document
//>`root_scope` The top-level `Scope` representing the document
//-> always 'true' (TODO: should actually return false on error)
bool docra::Parser_Source::parse(int verb, const fs::path &path, Scope &root_scope)
{
  Scope *curr_scope = &root_scope;
  Param *curr_param = nullptr;
  Suffix *curr_suffix = nullptr;
  Suffix *block_suffix = nullptr;
  Tag prefix_tag;
  int lineno_code = 0;
  enum {
    BLOCK_NONE, //Not currently parsing a block
    BLOCK_CODE, //Just a generic code block
    BLOCK_SIGN, //Block illustrating the signature of the current entity (unused!)
    BLOCK_PARAM //Block enclosing several interrelated params
  } curr_block_type = BLOCK_NONE;
  Tokens curr_block;
  Tokens backtrack;

  while (lexer.next_token()) {
    Token token (lexer.current_token());
    if (verb >= Config::VERB_INTERNAL) {
      std::cerr << token.lineno << ";";
    }
    if (token.depth < 0) {
      std::cerr << lexer.path.c_str() << ":" << token.lineno << " error: token.depth < 0" << std::endl;
      return false;
    }
    int leveloffset = 0;
    switch (token.kind) {
      case Token::TAG:
        backtrack.push_back(token);
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "TAG ("
            << token.tag.prefix << ", " << token.tag.symbol << ", " << token.tag.suffix << ")" << std::endl;
        }
        //TODO! apparently a TODO TAG inside a scope will override the current suffix! So that any subsequent '?' comment is appended to the TODO tag and not to the actual scope it's in!

        if (token.boc) {
          Param *p = make_param(curr_suffix, token);
          if (p) {
            curr_param = p;
            if (token.tag.suffix == "+") {
              curr_param->inlined = true;
              Token inline_param;
              inline_param.kind = Token::INLINE_PARAM;
              inline_param.lineno = token.lineno;
              inline_param.tag = token.tag;
              curr_suffix->desc_long.push_back(inline_param);
            }
            continue;
          }

          if (token.tag.prefix == "#") {
            curr_suffix = project.insert_tag(token.tag, std::move(token), nullptr);
            curr_suffix->path = path;
            // curr_suffix->line_code will be 0 by construction
            continue;
          } else {
            if (token.tag.suffix == "+") {
              if (token.tag.prefix == "+") {
                // Concatenate prefix tags
                prefix_tag.symbol += token.tag.symbol;
              } else {
                // Just replace previous prefix tag
                prefix_tag.erase();
                prefix_tag = token.tag;
              }
              continue;
            }
            // TODO TAGS SHOULD NOT OVERRIDE THE CURRENT PREFIX!!!!!
            Tag expanded_tag(token.tag);
            //? First tag inside a scope that inherited the tag from its parent, if we are on the same line as the scope opening then we are the actual tag for this scope!
            if (curr_scope && curr_scope->suffix && curr_scope->suffix->lineno_to() < curr_scope->lineno_start - 1
             && token.lineno == curr_scope->lineno_start) {
              if (curr_scope->parent && curr_scope->parent->suffix) {
                leveloffset = expand_tag(curr_scope->parent->suffix->tag, expanded_tag);
              } else if (!prefix_tag.empty()) {
                leveloffset = expand_tag(prefix_tag, expanded_tag);
              }
              curr_suffix = project.insert_tag(expanded_tag, std::move(token), curr_scope);
              curr_suffix->path = path;
              curr_suffix->lineno_code = lineno_code;
              curr_suffix->leveloffset = leveloffset;
              curr_scope->suffix = curr_suffix;
            } else if (curr_scope && !curr_scope->suffix && curr_scope->lineno_start >= token.lineno - 2) {
            //? First tag inside a scope that does not have one yet assigned, is assigned to the scope itself (if within a couple of lines from the start)
              if (curr_scope->parent && curr_scope->parent->suffix) {
                leveloffset = expand_tag(curr_scope->parent->suffix->tag, expanded_tag);
              } else if (!prefix_tag.empty()) {
                leveloffset = expand_tag(prefix_tag, expanded_tag);
              }
              curr_suffix = project.insert_tag(expanded_tag, std::move(token), curr_scope);
              curr_suffix->path = path;
              curr_suffix->lineno_code = lineno_code;
              curr_suffix->leveloffset = leveloffset;
              curr_scope->suffix = curr_suffix;
            } else if ((lineno_code >= token.lineno - 1) && (!curr_suffix
             || (std::max(curr_suffix->lineno_to(), curr_suffix->lineno_code) < lineno_code))) {
            //? If this tag follows some code, AND that code is not already tagged,
            //? then consider this tag to belong to that line of code. _
              if (curr_scope && curr_scope->suffix) {
                leveloffset = expand_tag(curr_scope->suffix->tag, expanded_tag);
              } else if (!prefix_tag.empty()) {
                leveloffset = expand_tag(prefix_tag, expanded_tag);
              }
              curr_suffix = project.insert_tag(expanded_tag, std::move(token), curr_scope);
              curr_suffix->path = path;
              curr_suffix->lineno_code = lineno_code;
              curr_suffix->leveloffset = leveloffset;
            } else {
            //? This tag does NOT have some code already associated with it, and it's not a '#' tag.
            // Then the code will come later, but we must already create a suffix for it!
            // (Or we have nowhere to put any subsequent `DOC_STR` s). _
              if (curr_scope && curr_scope->suffix) {
                leveloffset = expand_tag(curr_scope->suffix->tag, expanded_tag);
              } else if (!prefix_tag.empty()) {
                leveloffset = expand_tag(prefix_tag, expanded_tag);
              }
              curr_suffix = project.insert_tag(expanded_tag, std::move(token), curr_scope);
              curr_suffix->path = path;
              curr_suffix->leveloffset = leveloffset;
              // curr_suffix->line_code will be 0 by construction
            }
          }
        } else { //!token.boc
          if (curr_param) curr_param->desc.push_back(token);
          else if (curr_block_type != BLOCK_NONE) curr_block.push_back(token);
          else if (curr_suffix && token.lineno - 1 <= curr_suffix->lineno_to()) {
            if (!curr_suffix->desc_long.empty()) {
              curr_suffix->desc_long.push_back(token);
            } else {
              curr_suffix->desc_short.push_back(token);
            }
          }
        }
        break;
      case Token::SCOPE_ENTER:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "SCOPE_ENTER" << std::endl;
        }
        curr_scope = curr_scope->add_child(token.lineno);

//? If the scope starts on a line that was tagged, and that tag is NOT already matched, use that tag for the scope itself. (Add exception in case where tag is on the SAME line as the scope, because in this case the two clearly belong together). _
        if (curr_suffix && curr_suffix->lineno_code >= token.lineno - 2) {
          curr_scope->suffix = curr_suffix;
        }
        break;
      case Token::SCOPE_EXIT:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "SCOPE_EXIT" << std::endl;
        }

        lineno_code = curr_scope->lineno_start;
        // This is to prevent the current line (which is where the scope was closed)
        // to be tagged as code by a possible tag on the next line
        curr_scope->lineno_end = token.lineno;
        curr_scope = curr_scope->parent ? curr_scope->parent : &root_scope;
        if (curr_scope->suffix) curr_suffix = curr_scope->suffix;
        // Else keep the old suffix around, this should not hurt and will actually be helpful
        // if there's a BLOCK_EXIT somewhere next that would not have a suffix otherwise.
        break;
      case Token::DOC_STR:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "DOC_STR (" << token.str << ")" << std::endl;
        }

        //? A DOC_STR inside a block is just added verbatim to that block. _
        if (curr_block_type != BLOCK_NONE) {
          curr_block.push_back(token);
          continue;
        }

        //? If a comment starts with a tag, it starts a tagged block
        // Comments on lines immediately following it should be considered part of the same block.
        // (but they must start with a space! this is a simple way to separate them from commented code). _

        if (token.boc) {
          if (token.str.length() > 1 && token.str.substr(0, 2) == "? ") {
            //? If a "? " comment occurs inside a block, it will be rendered in asciidoc as a numbered code block callout.
            // (but all I have to do here is add it to the description, the renderer will do the rest). _
            if (curr_block_type != BLOCK_NONE) {
              curr_block.push_back(token);
              continue;
            } else if (!curr_scope->suffix || curr_suffix != curr_scope->suffix) {
            //? If a comment starts with "? ", it starts a hierachical block (or it continues the previous one, there is no difference)
            // As such it should associate with the innermost containing scope with a suffix associated to it. _
              // Recurse up the scope stack to find where this belongs..
              Scope *suffix_scope = curr_scope;
              do {
                if (suffix_scope->suffix) {
                  curr_suffix = suffix_scope->suffix;
                  break;
                } else {
                  suffix_scope = suffix_scope->parent;
                }
              } while (suffix_scope);
            }
          } else {
          //? Comments that end up here did NOT start with "? ", they will be accepted as docstrings only if
          // they are preceded by a line belonging to a documentation block. _
            if (!curr_suffix || token.lineno - 1 > curr_suffix->lineno_to()) {
              backtrack.push_back(token);
              continue;
            }
            //? Additional conditions for valid comments are that: _
            // Either they are empty (and will count as an empty line in the resulting asciidoc). _
            if (token.empty_str()) {
              // Do nothing, and proceed to add the comment
            } else if (token.str.substr(0, 2) == "->") {
            //? - or they start with an arrow sign, in which case they are documenting a function return value. _
              curr_suffix->params.emplace_back(Param::RET, "", token.lineno);
              curr_param = &curr_suffix->params.back();
            } else if (token.str.front() == ' ') {
            //? - or they must start with a space, as a way to distinguish them from a commented line of code. _
              //Do nothing, and proceed to add the comment
            } else {
              // Skip this comment
              backtrack.push_back(token);
              continue;
            }
          }
        } else if ((!curr_suffix || curr_suffix->lineno_to() != token.lineno) && !curr_param) {
        // For comment strings that occurr after the first line we have to remember if this line must be ignored or not!
          backtrack.push_back(token);
          continue;
        }
        if (curr_param) {
          curr_param->desc.push_back(token);
        } else if (curr_block_type != BLOCK_NONE) {
          curr_block.push_back(token);
        } else if (curr_suffix) {
          //curr_suffix->desc.push_back(token);
          if (!curr_suffix->desc_long.empty()
           || (!curr_suffix->desc_short.empty() && !curr_suffix->desc_code.empty())
           || (token.boc && token.str.length() > 1 && token.str.substr(0, 2) == "? ")) {
            curr_suffix->desc_long.push_back(token);
          } else {
            curr_suffix->desc_short.push_back(token);
          }
        }
        //else the comment is just ignored
        break;
      case Token::CODE:
        backtrack.push_back(token);
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "CODE (" << token.str << ")" << std::endl;
        }
        if (!token.empty_str()) {
          lineno_code = token.lineno;

          // Find whether this code is following an unassigned tag, and if so assign it
          if (curr_suffix && curr_suffix->lineno_code == 0 && curr_suffix->lineno_to() >= lineno_code - 1) {
            curr_suffix->lineno_code = lineno_code;
          }
        }

        // If we're inside a block, add tokens to the suffix descriptions
        if (curr_block_type != BLOCK_NONE) {
          curr_block.push_back(token);
        }
        break;
      case Token::BLOCK_ENTER:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "BLOCK_ENTER" << std::endl;
        }
        if (curr_block_type != BLOCK_NONE) {
          error(path, token, "Nested code blocks are not allowed!");
          return false;
        }
        if (curr_suffix) {
          curr_block_type = BLOCK_CODE;
          curr_block.push_back(token);
          block_suffix = curr_suffix;
        } else {
          error(path, token, "Opening a block without an associated suffix is an error");
          return false;
        }
        break;
      case Token::BLOCK_EXIT:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "BLOCK_EXIT" << std::endl;
        }
        if (!block_suffix) {
          block_suffix = curr_suffix;
        }
        if (!block_suffix) {
          //TODO: this is not a serious error and should not cause the parser to give up
          error(path, token, "BLOCK_EXIT token without an associated tag!");
          return false;
        }

        //TODO: ensure the backtrack buffer is not empty?
        if (curr_block_type == BLOCK_NONE) {
          int lineno_to = block_suffix->lineno_to();
          if (lineno_to == token.lineno) {
            --lineno_to; //Make it accept that line only
          }
          //If the backtrack starts on an empty line of code, we should forget about this line and jump back by one)
          auto i = backtrack.rbegin();
          if (i->kind == Token::CODE && i->empty_str()) {
            --lineno_to;
          }
          for (auto i = backtrack.rbegin(); i != backtrack.rend(); ++i) {
            if (i->lineno <= lineno_to) {
              break;
            }
            if (i->kind == Token::TAG && i->boc) {
              continue; //boc tags do not belong here
            }
            curr_block.push_front(*i);
          }
          backtrack.clear();

          Token fake_enter;
          fake_enter.kind = Token::BLOCK_ENTER;
          fake_enter.lineno = lineno_to;
          fake_enter.str = "<?";
          curr_block.push_front(fake_enter);
        }
        curr_block.push_back(token);
        //block_suffix->desc.splice(block_suffix->desc.end(), curr_block);

        if (curr_param) {
          if (curr_param->desc_code.empty()) {
            curr_param->desc_code.splice(curr_param->desc_code.end(), curr_block);
          } else {
            curr_param->desc.splice(curr_param->desc_code.end(), curr_block);
          }
        } else if (block_suffix->desc_code.empty() && block_suffix->desc_long.empty()) {
          block_suffix->desc_code.splice(block_suffix->desc_code.end(), curr_block);
        } else {
          block_suffix->desc_long.splice(block_suffix->desc_long.end(), curr_block);
        }
        curr_block_type = BLOCK_NONE;
        block_suffix = nullptr;
        break;
      case Token::LITERAL:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::string(token.depth, ' ') << "LITERAL(" << token.str << ")" << std::endl;
        }
#ifndef NDEBUG
        // A comment literal with this magic string can be used to set breakpoints at arbitrary points in the input stream
        if (token.str == "__DOCRA_BREAK__") {
          continue;
        }
#endif
        if (curr_block_type != BLOCK_NONE) {
          curr_block.push_back(token);
        } else if (curr_suffix) {
          if (curr_param) {
            curr_param->desc.push_back(token);
          } else if (!curr_suffix->desc_long.empty()) {
            curr_suffix->desc_long.push_back(token);
          } else {
            curr_suffix->desc_short.push_back(token);
            //curr_suffix->desc.push_back(token);
          }
        }
        break;
      case Token::EOL:
        backtrack.push_back(token);
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::endl;
        }
        if (curr_param) {
          curr_param = nullptr;
        } else if (curr_block_type != BLOCK_NONE) {
          curr_block.push_back(token);
        } else if (curr_suffix && curr_suffix->lineno_to() == token.lineno) {
          if (!curr_suffix->desc_long.empty()) {
            curr_suffix->desc_long.push_back(token);
          } else if (!curr_suffix->desc_short.empty()) {
            curr_suffix->desc_short.push_back(token);
          }
          //curr_suffix->desc.push_back(token);
        }
        break;
    }
  }
  if (verb >= Config::VERB_DEBUG) {
    root_scope.print();
  }
  return true;
}

//docra`Suffix`+

//++`::lineno_to`
int docra::Suffix::lineno_to()
{
  int lineno_desc_short = desc_short.empty() ? lineno_from() : desc_short.back().lineno;
  int lineno_desc_code = desc_code.empty() ? lineno_desc_short : desc_code.back().lineno;
  int lineno_desc_long = desc_long.empty() ? lineno_desc_code : desc_long.back().lineno;
  int lineno_desc = std::max(lineno_desc_code, std::max(lineno_desc_short, lineno_desc_long));
  int lineno_para = params.empty() ? lineno_from()
   : params.back().desc.empty() ? params.back().lineno : params.back().desc.back().lineno;
  return std::max(lineno_desc, lineno_para);
}

void docra::Suffix::render_pilcrow(const Config &cfg, std::ostream &out, int lineno) const
{
  if(!cfg.cli_urlprefix.empty()) {
    out << std::endl
      << "ifeval::[\"{basebackend}\" == \"html\"]" << std::endl
      << "ifdef::linkattrs[]" << std::endl
      << cfg.cli_urlprefix << path.generic_string() << cfg.cli_urlsuffix << lineno << "[&#182;, role=\"pilcrow\"]" << std::endl
      << "endif::[]" << std::endl
      << "ifndef::linkattrs[]" << std::endl
      << cfg.cli_urlprefix << path.generic_string() << cfg.cli_urlsuffix << lineno << "[&#182;]" << std::endl
      << "endif::[]" << std::endl
      << "endif::[]";
  }
  //out << std::endl;
}

bool docra::Suffix::render_description(const Config &cfg, std::ostream &out, const Tokens &desc) const
{
  Tag rtag;
  int lineno = -1;
  bool ispar = false;
  auto it = desc.begin();
  while (it != desc.end()) {
    auto in = it; if (in != desc.end()) ++in;
    auto inn = in; if (inn != desc.end()) ++inn;
    bool endpar = in == desc.end() || (lineno > 0 && in->lineno > lineno + 1) || (in->kind == Token::EOL && inn == desc.end());
    switch (it->kind) {
      case Token::EOL:
        if (ispar && endpar) { render_pilcrow(cfg, out, it->lineno); ispar = false; }
        out << std::endl;
        break;
      case Token::TAG:
        if (resolve(it->tag, rtag)) {
          out << "<<" << rtag.xref_id() << ">>";
        } else {
          out << "``" << it->tag.prefix << "`" << it->tag.symbol << "`" << it->tag.suffix << "``";
          std::cerr << path.c_str() << ":" << it->lineno << " warning: unable to resolve tag "
           << it->tag.prefix << "`" << it->tag.symbol << "`" << it->tag.suffix << std::endl;
        }
        ispar = true;
        break;
      case Token::INLINE_PARAM: {
        const Param *param = find_param(it->tag.symbol);
        if (!param) {
          //TODO: when an inline param does not resolve, a warning should be emitted
          out << "_ " << it->tag.symbol << "::" << std::endl;
          out << "_Unresolved param_" << std::endl;
        } else {
          param->render(cfg, *this, out);
        }
        break;
      }
      case Token::DOC_STR: {
        auto from = it->str.begin();
        auto to = it->str.end();
        bool newline = false;

        if (it == desc.begin() && it->empty_str()) {
          break;
        }

        // If this str ends in " _" AND is followed by an EOL, then insert a newline in the output
        if ((in == desc.end() || in->kind == Token::EOL)
         && it->str.length() >= 2 && it->str.substr(it->str.length() - 2, 2) == " _") {
          to -= 1;
          newline = true;
        }
        if (it->boc) {
          // If this str begins in "++=" Then this is a heading (relative to the current leveloffset)
          if (it->str.length() >= 2 && it->str.substr(0, 2) == "? ") {
            from += 1;
          }
          if ((to - from >= 2) && std::string(from, from + 2) == " _") {
            from += 2;
            // If this line starts with a " _" then insert a newline before anything else
            out << std::endl;
          }
          if ((to - from >= 4) && std::string(from, from + 4) == " ++="){
            from += 3;
            out << std::endl << std::endl;
            out << std::string(leveloffset + 1, '=');
          }
          // If the first char is a space, skip it or asciidoc will emit a literal block
          if ((to - from >= 1) && *from == ' ') {
            from += 1;
          }
        }
        if (from != to) {
          ispar = true; // Only add a pilcrow when it's a non-empty line
        }
        out << std::string(from, to);
        if (ispar && endpar) { render_pilcrow(cfg, out, it->lineno); ispar = false; }
        if (newline) out << std::endl;
        break;
      }
      case Token::BLOCK_ENTER:
        render_codeblock(out, desc, it);
        if (it != desc.end()) {
          in = ++it;
        }
        break;
      case Token::LITERAL:
        out << "`" << it->str << "`";
        ispar = true;
        break;
      default:
        out << it->str;
        ispar = true;
        break;
    }
    if (in != desc.end()) lineno = in->lineno;
    it = in;
  }
  return true;
}

bool docra::Suffix::render_codeblock(std::ostream &out, const Tokens &desc, docra::Tokens::const_iterator &it) const
{
  if (it->kind != Token::BLOCK_ENTER) {
    return false;
  }
  // Step 1: loop through all the lines to calculate indentation

  // Step 2: actually output the code block

  std::list<std::list<const Token *>> callouts;

  out << std::endl
    << "[source,indent=0]" << std::endl
    << "----" << std::endl;

  while (++it != desc.cend()) {
endcallout:
    switch (it->kind) {
      case Token::BLOCK_EXIT:
        goto endblock;
      case Token::CODE:
        out << it->str;
        break;
      case Token::DOC_STR:
        if (it->str.substr(0, 2) == "? ") {
          callouts.emplace_back();
          auto &callout = callouts.back();
          do {
            switch (it->kind) {
              default:
                callout.push_back(&(*it));
                break;
              case Token::EOL:
                out << " // <" << callouts.size() << "> ";
                goto endcallout;
            }
          } while (++it != desc.end());
        } else {
        // Else this comment is not documentation but just commented-out code
        // (We'll still print it, for now)
          out << it->str;
        }
        break;
      case Token::EOL:
        out << std::endl;
        break;
      default:
        std::cerr << path.c_str() << ":" << it->lineno << " warning: unexpected token in codeblock, kind = " << it->kind << std::endl;
        break;
    }
  }
endblock:

  out << std::endl << "----" << std::endl;

  // Generate backreferences, if any
  Tag rtag;
  int co = 1;
  for (auto &callout : callouts) {
    out << "<" << co << "> ";
    int i = 0;
    for (auto token : callout) {
      switch (token->kind) {
        case Token::DOC_STR:
          if (i == 0 && token->str.substr(0, 2) == "? ") {
            out << token->str.substr(2);
          } else {
            out << token->str;
          }
          break;
        case Token::TAG:
          if (resolve(token->tag, rtag)) {
            out << "<<" << rtag.xref_id() << ">>";
          } else {
            out << "``" << token->tag.prefix << "`" << token->tag.symbol << "`" << token->tag.suffix << "``";
            std::cerr << path.c_str() << ":" << token->lineno << " warning: unable to resolve tag "
             << token->tag.prefix << "`" << token->tag.symbol << "`" << token->tag.suffix << std::endl;
          }
          break;
        case Token::LITERAL:
          out << "`" << token->str << "`";
          break;
        default:
          std::cerr << path.c_str() << ":" << token->lineno << " warning: unexpected token in callout, kind = " << token->kind << std::endl;
          break;
      }
      ++i;
    }
    out << std::endl;
    ++co;
  }
  out << std::endl;

  return true;
}

bool docra::Suffix::render(const Config &cfg, std::ostream &out) const
{
  render_description(cfg, out, desc_short);
  if (!desc_code.empty()) {
    auto ci = desc_code.begin();
    render_codeblock(out, desc_code, ci);
  }
  render_description(cfg, out, desc_long);
  return true;
}

//++`.resolve`
docra::Suffix *docra::Suffix::resolve(const Tag &in, Tag &out) const
{
  out = in;

  Suffix *ret = project->locate(in);
  if (ret) return ret;

  // first locate the prefix
  Prefix *prefix = project->locate_prefix(in.prefix);
  if (!prefix) return nullptr;

  for (auto &sym : prefix->symbols) {
    out.symbol = sym.first + in.symbol;
    Suffix *resolved = project->locate(out);
    if (resolved) return resolved;
  }

  //TODO:if a perfect match was not found, try to match at word boundaries instead of using the full string

  out.symbol = in.symbol;
  return nullptr;
}

const docra::Param *docra::Suffix::find_param(const std::string &name) const
{
  for (const auto &param : params) {
    if (param.name == name) {
      return &param;
    }
  }
  return nullptr;
}

bool docra::Symbol::render_xref(const Config &cfg, std::ostream &out) const
{
  if (suffixes.empty()) {
    return true;
  }

  int leveloffset = 0;
  for (auto &suffix : suffixes) {
    leveloffset = std::max(leveloffset, suffix.second.leveloffset);
  }

  const Tag& tag = suffixes.begin()->second.tag;
  if (suffixes.size() == 1 && tag.suffix == "+") {
    // Ignore this as it's only used to compose longer symbols
    // In fact, it should not be here at all! TODO!
    return true;
  }
  //if (tag.extra) {
  //  tag.extra->render_xref();
  //} else {
    out << "[[" << tag.xref_id() << "]]" << std::endl;
    out << "===" << std::string(leveloffset, '=');
    out << " `" << tag.symbol << "`" << std::endl << std::endl;
  //}

  for (auto range = suffixes.equal_range(suffixes.begin()->first);;
   range = suffixes.equal_range(range.second->first)) {
    // Render the short description, if any
    auto it = range.first;
    do {
      it->second.render_description(cfg, out, it->second.desc_short);
      out << std::endl;
    } while (++it != range.second);
    // Render the code block, if any
    it = range.first;
    do {
      if (!it->second.desc_code.empty()) {
        auto ci = it->second.desc_code.begin();
        it->second.render_codeblock(out, it->second.desc_code, ci);
      }
      out << std::endl;
    } while (++it != range.second);
    int locals = 0;
    int returns = 0;
    // Render the params, if any
    it = range.first;
    do {
      for (auto param : it->second.params) {
        if (param.inlined) continue;
        if (param.kind == Param::RET) {
          ++returns; continue;
        }
        param.render(cfg, it->second, out);
        out << std::endl;
      }
      out << std::endl;
    } while (++it != range.second);
    // Render the return type, if any
    if (returns) {
      it = range.first;
      do {
        for (auto param : it->second.params) {
          if (param.inlined) continue;
          switch (param.kind) {
            default:
              continue;
            case Param::RET:
              break;
          }
          out << std::endl << "*Returns* ";
          it->second.render_description(cfg, out, param.desc);
          out << std::endl;
        }
        out << std::endl;
      } while (++it != range.second);
    }
    // Render the local variables, if any
    if (locals) {
      out << std::endl << "====" << std::string(leveloffset, '=');
      out << " Locals" << std::endl;
      it = range.first;
      do {
        for (auto param : it->second.params) {
          if (param.inlined) continue;
          switch (param.kind) {
            case Param::LOCAL:
              out << "`" << param.name << "`::" << std::endl;
              break;
            default:
              continue;
          }
          it->second.render_description(cfg, out, param.desc);
          out << std::endl;
        }
        out << std::endl;
      } while (++it != range.second);
    }
    // Render the source locations
    it = range.first;
    do {
      auto path = it->second.path.generic_string();
      auto line = it->second.lineno_code ? it->second.lineno_code : it->second.lineno_from();
      if (cfg.cli_urlprefix.empty()) {
        out << path << ":" << line << std::endl << std::endl;
      } else {
        out << cfg.cli_urlprefix << path << cfg.cli_urlsuffix << line
         << "[" << path << ":" << line << "]" << std::endl << std::endl;
      }
      out << std::endl;
    } while (++it != range.second);
    // Render the long description, if any
    it = range.first;
    do {
      it->second.render_description(cfg, out, it->second.desc_long);
      out << std::endl;
    } while (++it != range.second);
    out << "'''" << std::endl;
    if (range.second == suffixes.end()) break;
  }
  return true;
}

bool docra::Symbol::render_xref_yacc(const Config &cfg, std::ostream &out) const
{
  if (suffixes.empty()) {
    return true;
  }

  int leveloffset = 0;
  for (auto &suffix : suffixes) {
    leveloffset = std::max(leveloffset, suffix.second.leveloffset);
  }

  int pattern_footnote_idx = 0;
  for (auto &suffix : suffixes) {
    const Tag& tag = suffix.second.tag;
    if (suffix.first.empty()) {
      out << "[[" << tag.xref_id() << "]]" << std::endl;
      out << "===" << std::string(leveloffset, '=');
      out << " `" << tag.symbol << "`" << std::endl << std::endl;
      suffix.second.render_description(cfg, out, suffix.second.desc_short);
      out << std::endl;

      // Render the source location
      auto path = suffix.second.path.generic_string();
      auto line = suffix.second.lineno_code ? suffix.second.lineno_code : suffix.second.lineno_from();
      if (cfg.cli_urlprefix.empty()) {
        out << path << ":" << line << std::endl << std::endl;
      } else {
        out << cfg.cli_urlprefix << path << cfg.cli_urlsuffix << line
        << "[" << path << ":" << line << "]" << std::endl << std::endl;
      }
      out << std::endl;
    } else if (suffix.second.token_tag.extra
     && suffix.second.token_tag.extra->kind == ExtraInfo::YACC) {
      if (tag.suffix == "0") {
        out << "`:` ";
      } else {
        out << "`|` ";
      }
      auto extra = static_cast<ExtraInfo_Yacc *>
       (suffix.second.token_tag.extra.get());
      Tag xref_tag;
      xref_tag.prefix = tag.prefix;
      for (auto &rule : extra->pattern) {
        if (rule.first.empty()) {
          out << "_nothing_ ";
          continue;
        }
        xref_tag.symbol = rule.second;
        out << "<<" << xref_tag.xref_id() << ",`+++" << rule.first << "+++`>> ";
      }
      // If this pattern has a description, append a progressive number (used for crossreference)
      if (!suffix.second.desc_short.empty() || !suffix.second.desc_long.empty()) {
        ++pattern_footnote_idx;
        out << "^*[" << pattern_footnote_idx << "]*^ ";
      }
      out << "+"; // Append explicit line break
    }
    out << std::endl;
  }
  if (suffixes.size() > 1) {
    out << "`;`" << std::endl;
  }

  // Render footnotes
  pattern_footnote_idx = 0;
  for (auto &suffix : suffixes) {
    if (suffix.first.empty()) {
      continue;
    }
    if (suffix.second.desc_short.empty() && suffix.second.desc_long.empty()) {
      continue;
    }
    ++pattern_footnote_idx;
    out << std::endl << "*[" << pattern_footnote_idx << "]* ";
    suffix.second.render(cfg, out);
    out << std::endl;
  }

  auto rule_name = suffixes.begin();
  out << std::endl;
  rule_name->second.render_description(cfg, out, rule_name->second.desc_long);
  return true;
}

//docra`Parser_Template`+
docra::Parser_Template::Parser_Template(Lexer_Adoc &l, Project &p, std::ostream &o)
: lexer(l), project(p), out(o)
{}

//++`::parse`
bool docra::Parser_Template::parse(int verb)
{
  static std::regex re_directive(
    "\\?([_[:alpha:]]\\w*)(\\[([^\\]]+)\\])?(:)?\\s*"
  );
  enum {
    RE_MATCH = 0,
    RE_ID,
    RE_SQBRAKETS,
    RE_OPTS,
    RE_COLON
  };
  std::smatch match;

  while (next_token()) {
    const Token &token = current_token();
    if (verb >= Config::VERB_INTERNAL) {
      std::cerr << token.lineno << ";";
    }
    switch (token.kind) {
      case Token::TAG: {
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << token.str;
        }
        Suffix *suffix = project.locate(token.tag);
        if (suffix == nullptr) {
          std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: referenced tag not found: " << token.tag << std::endl;
          continue;
        }
        Tokens tokens;
        tokens.push_back(token);
        suffix->render_description(project._cfg, out, tokens);
        break;
      }
      case Token::DOC_STR:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << token.str;
        }
        if (token.boc) {
          if (!std::regex_search(token.str, match, re_directive)) {
            out << "//" << token.str;
          } else {
            std::map<std::string, std::string> directive_args;
            if (match[RE_OPTS].matched) {
              static std::regex re_opt(
                "^\\s*([_[:alpha:]][_[:alnum:]]*)\\s*(=\\s*([^,]+))?\\s*,?"
              );
              enum {
                RE_OPT_MATCH = 0,
                RE_OPT_ID,
                RE_OPT_HASVAL,
                RE_OPT_VAL
              };
              std::smatch opt_match;
              std::string opt_str = match[RE_OPTS];
              while (std::regex_search(opt_str, opt_match, re_opt)) {
                if (opt_match[RE_OPT_HASVAL].matched) {
                  directive_args[opt_match[RE_OPT_ID]] = opt_match[RE_OPT_VAL];
                } else {
                  directive_args[opt_match[RE_OPT_ID]] = "";
                }
                std::cerr << opt_match[RE_OPT_ID] << "=" << directive_args[opt_match[RE_OPT_ID]] << std::endl;
                opt_str = opt_match.suffix();
              }
            }
            if (match[RE_ID] == "includetag") {
              if (!match[RE_COLON].matched) {
                std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: missing ':' after directive" << std::endl;
              } else {
                // Get the next token and make sure it's a tag
                if (next_token()) {
                  const Token &tagtok = current_token();
                  if (tagtok.kind != Token::TAG) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: missing tag after includetag directive" << std::endl;
                    continue;
                  }
                  Suffix *suffix = project.locate(tagtok.tag);
                  if (suffix == nullptr) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: included tag not found: " << tagtok.tag << std::endl;
                    continue;
                  }
                  suffix->render(project._cfg, out);
                }
                if (next_token()) {
                  const Token &tageol = current_token();
                  if (tageol.kind != Token::EOL) {
                    // do what?
                  }
                  // Do not print it to avoid an extra EOL after the description
                }
              }
            } else if (match[RE_ID] == "includetag_sorted") {
              if (!match[RE_COLON].matched) {
                std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: missing ':' after directive" << std::endl;
              } else {
                if (next_token()) {
                  const Token &tagtok = current_token();
                  if (tagtok.kind != Token::TAG) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: missing tag after includetag_unsorted directive" << std::endl;
                    continue;
                  }
                  Symbol *symbol = project.locate_symbol(tagtok.tag);
                  if (symbol == nullptr) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: included tag not found: " << tagtok.tag << std::endl;
                    continue;
                  }
                  for (auto &suffix : symbol->suffixes) {
                    suffix.second.render(project._cfg, out);
                  }
                }
                if (next_token()) {
                  const Token &tageol = current_token();
                  if (tageol.kind != Token::EOL) {
                    // do what?
                  }
                  // Do not print it to avoid an extra EOL after the description
                }
              }
            } else if (match[RE_ID] == "includetag_unsorted") {
              if (!match[RE_COLON].matched) {
                std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: missing ':' after directive" << std::endl;
              } else {
                // Get the next token and make sure it's a tag
                if (next_token()) {
                  const Token &tagtok = current_token();
                  if (tagtok.kind != Token::TAG) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: missing tag after includetag_unsorted directive" << std::endl;
                    continue;
                  }
                  Symbol *symbol = project.locate_symbol(tagtok.tag);
                  if (symbol == nullptr) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: included tag not found: " << tagtok.tag << std::endl;
                    continue;
                  }
                  for (auto &suffix : symbol->suffixes_unsorted) {
                    suffix->second.render(project._cfg, out);
                  }
                }
                if (next_token()) {
                  const Token &tageol = current_token();
                  if (tageol.kind != Token::EOL) {
                    // do what?
                  }
                  // Do not print it to avoid an extra EOL after the description
                }
              }
            } else if (match[RE_ID] == "includexref") {
              if (!match[RE_COLON].matched) {
                Prefix *prefix = project.locate_prefix("");
                if (prefix == nullptr) {
                  std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: could not locate prefix \"\"" << std::endl;
                  continue;
                }
                for (auto &symbol : prefix->symbols) {
                  symbol.second.render_xref(project._cfg, out);
                }
              } else {
                if (next_token()) {
                  const Token &tagtok = current_token();
                  if (tagtok.kind != Token::TAG) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: missing tag after includexref directive" << std::endl;
                    continue;
                  }
                  std::regex re_prefix = make_regex(tagtok.tag.prefix);
                  std::regex re_symbol = make_regex(tagtok.tag.symbol);

                  for (auto &prefix : project.prefixes) {
                    if (!std::regex_match(prefix.first, re_prefix)) {
                      continue;
                    }
                    for (auto &symbol : prefix.second.symbols) {
                      if (!std::regex_match(symbol.first, re_symbol)) {
                        continue;
                      }
                      symbol.second.render_xref(project._cfg, out);
                    }
                  }
                }
              }
            } else if (match[RE_ID] == "includexref_yacc") {
              if (!match[RE_COLON].matched) {
                Prefix *prefix = project.locate_prefix("");
                if (prefix == nullptr) {
                  std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: could not locate prefix \"\"" << std::endl;
                  continue;
                }
                for (auto &symbol : prefix->symbols) {
                  symbol.second.render_xref_yacc(project._cfg, out);
                }
              } else {
                if (next_token()) {
                  const Token &tagtok = current_token();
                  if (tagtok.kind != Token::TAG) {
                    std::cerr << lexer.path.c_str() << ":" << tagtok.lineno << " warning: missing tag after includexref_yacc directive" << std::endl;
                    continue;
                  }
                  std::regex re_prefix = make_regex(tagtok.tag.prefix);
                  std::regex re_symbol = make_regex(tagtok.tag.symbol);

                  for (auto &prefix : project.prefixes) {
                    if (!std::regex_match(prefix.first, re_prefix)) {
                      continue;
                    }
                    for (auto &symbol : prefix.second.symbols) {
                      if (!std::regex_match(symbol.first, re_symbol)) {
                        continue;
                      }
                      symbol.second.render_xref_yacc(project._cfg, out);
                    }
                  }
                }
              }
            } else {
              out << "//" << token.str;
            }
          }
        } else {
          out << token.str;
        }
        break;
      case Token::CODE:
        out << token.str;
        break;
      case Token::LITERAL:
        out << "`" << token.str << "`";
        break;
      case Token::EOL:
        if (verb >= Config::VERB_INTERNAL) {
          std::cerr << std::endl;
        }
        out << std::endl;
        break;
      default:
        std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: unexpected token in template, kind = " << token.kind << std::endl;
        break;
    }
  }
  return true;
}

bool docra::Parser_Template::next_token()
{
  static std::regex re_tag(
   "``([^`]+)``" "|"
   "(([^\\s`]*)`([^`]*)`([^\\s:]*))(:)?"
  );
  enum {
    RE_MATCH = 0,
    RE_LITERAL,
    RE_TAG, RE_TAG_PREFIX, RE_TAG_SYMBOL, RE_TAG_SUFFIX, RE_TAG_COLON,
  };

  if (nested_token.kind == Token::EOL
   || it == lexer.current_token().str.cend()) {
    if (!lexer.next_token()) {
      return false;
    }
    it = lexer.current_token().str.cbegin();
  }
  const Token &token = lexer.current_token();
  if (token.kind != Token::CODE) {
    nested_token.kind = Token::EOL;
    return true;
  }

matched:

  if (match.size() > 1) {
    if (match[RE_LITERAL].matched) {
      nested_token.kind = Token::LITERAL;
      nested_token.str = match[RE_LITERAL];
    } else if (match[RE_TAG].matched) {
      nested_token.kind = Token::TAG;
      nested_token.str = match[RE_TAG];
      nested_token.tag.prefix = match[RE_TAG_PREFIX];
      nested_token.tag.symbol = match[RE_TAG_SYMBOL];
      nested_token.tag.suffix = match[RE_TAG_SUFFIX];
      nested_token.brief = match.length(RE_TAG_COLON) == 0;
    }
    nested_token.lineno = token.lineno;
    if (match.suffix().matched) {
      it = match.suffix().first;
    } else {
      it = lexer.current_token().str.cend();
    }
    std::smatch empty_match;
    swap(match, empty_match);
    return true;
  }

  if (std::regex_search(it, token.str.cend(), match, re_tag)) {
    if (match.prefix().matched) {
      nested_token.kind = Token::CODE;
      nested_token.str = match.prefix();
      nested_token.lineno = token.lineno;
      it += match.prefix().length();
      return true;
    } else {
      goto matched;
    }
  } else {
    nested_token.kind = Token::CODE;
    nested_token.str.assign(it, token.str.cend());
    nested_token.lineno = token.lineno;
    it = lexer.current_token().str.cend();
    return true;
  }
}

const docra::Token &docra::Parser_Template::current_token()
{
  const Token &token = lexer.current_token();
  if (nested_token.kind == Token::EOL
   || it == token.str.cbegin()) {
    return token;
  } else {
    return nested_token;
  }
}

//docra`Parser_Config`+

//ug`config.syntax` The configuration file format is very similar to https://git-scm.com/docs/git-config[git].
// Lines that start with a ``#`` or ``;`` character are ignored and can be used for comments (including docra documentation!).
// The rest of the file is composed of sections and variables.
// A section begins with the name of the section in square brackets: <?
// [sectionname]
//?>and ends when the next section begins or the file ends.
// In some cases, a section must be defined with an additional _subsection_ name in quotes, like so: <?
// [sectionname "subsectionname"]
//?>Any non-empty line that is not a section header is recognized as setting a variable, since each variable must belong to a section, the first variable in a section must be preceded by a section header. Variables follow the following format: <?
// variable = value
//?>
// The following sections and variables are recognized:

//++`::parse`
bool docra::Parser_Config::parse(const fs::path &path)
{
  static std::regex re_section("\\[\\s*(\\S*)(\\s+\"(.*)\")?\\s*\\]");
  enum {
    RE_SEC_MATCH = 0,
    RE_SECTION,
    RE_SUBSEC_QUOTE,
    RE_SUBSEC
  };

  static std::regex re_keyval("\\s*([\\S]+)\\s*=\\s*(.*)");
  enum {
    RE_KV_MATCH = 0,
    RE_KEY,
    RE_VALUE
  };

//ug`config.syntax`lexer.adoc.2 This lexer is also used to scan docra templates.
  Lexer_Ini lexer;
  FILE *pfile = fopen(path.string().c_str(), "r");
  if (pfile == nullptr) {
    return false;
  }
  lexer.init(path.string(), pfile);

  _cfg.reset();

  _cfg.base_dir = path.parent_path();
  if (_cfg.base_dir.empty()) {
    _cfg.base_dir = fs::current_path();
  }

  std::smatch match_sec, match_kv;
  std::string sec, sub;

  while (lexer.next_token()) {
    const Token &token = lexer.current_token();
    switch (token.kind) {
      case Token::CODE:
        if (token.empty_str()) {
          continue;
        }
        if (std::regex_match(token.str, match_sec, re_section)) {
          sec = match_sec[RE_SECTION];
          //std::cout << "[" << match_sec[RE_SECTION];
          if (match_sec[RE_SUBSEC].matched) {
          //  std::cout << " \"" << match_sec[RE_SUBSEC] << "\"";
            sub = match_sec[RE_SUBSEC];
          } else {
            sub.erase();
          }
          //std::cout << "]";
          _cfg.set(sec, sub);
        } else {
          std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: malformed config section" << std::endl;
        }
        break;
      case Token::SCOPE_ENTER:
        while (lexer.next_token()) {
          const Token &tokkv = lexer.current_token();
          if (token.kind == Token::SCOPE_EXIT) {
            break;
          }
          if (token.kind != Token::CODE) {
            continue;
          }
          if (token.empty_str()) {
            continue;
          }
          if (std::regex_match(tokkv.str, match_kv, re_keyval)) {
            //std::cout << match_kv[RE_KEY] << " = " << match_kv[RE_VALUE] << std::endl;
            std::string value = match_kv[RE_VALUE];
            _cfg.set(sec, sub, match_kv[RE_KEY], trim_right_spaces(value), false);
          } else {
            std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: malformed config param" << std::endl;
          }
        }
        break;
      case Token::SCOPE_EXIT:
        break;
      case Token::EOL:
        //std::cout << std::endl;
        break;
      case Token::TAG:
      case Token::DOC_STR:
      case Token::LITERAL:
      case Token::BLOCK_ENTER:
      case Token::BLOCK_EXIT:
        while (lexer.next_token()) {
          if (token.kind == Token::EOL)
            break;
        }
        break;
      default:
        std::cerr << lexer.path.c_str() << ":" << token.lineno << " warning: unexpected token in config, kind = " << token.kind << std::endl;
        break;
    }
  }
  return true;
}
